import types

from .. import _impl_column, _impl_specimen
from ...frozen_instrumentation import HereditaryStratigraphicSpecimen
from ...genome_instrumentation import HereditaryStratigraphicColumn


def dispatch_impl(first: object, second: object) -> types.ModuleType:
    """Return the implementation module corresponding to the input target.

    Parameters
    ----------
    first, second : object
        The target objects for which to obtain the implementation module.
        It can be an instance of HereditaryStratigraphicColumn or
        HereditaryStratigraphicSpecimen.

    Returns
    -------
    types.ModuleType
        The implementation module corresponding to the input target.

    Raises
    ------
    TypeError
        If the input target is not an instance of either
        HereditaryStratigraphicColumn or HereditaryStratigraphicSpecimen.
    """
    assert type(first) == type(second)
    if isinstance(first, HereditaryStratigraphicColumn):
        return _impl_column
    elif isinstance(first, HereditaryStratigraphicSpecimen):
        return _impl_specimen
    else:
        raise TypeError(first)



import numpy as np

from hstrat import hstrat


def test_estimate_rank_of_mrca_maximum_likelihood_no_common_ancestor1():
    coincident_ranks = iter([0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert result is None


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor1():
    coincident_ranks = iter([1, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert result == 0


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor2():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert np.isclose(result, 4.5)


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor3():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert np.isclose(result, 7)


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor4():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.UniformPrior()
    )

    assert np.isclose(result, 4)


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor5():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5**64

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert np.isclose(result, 7)


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor6():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.UniformPrior()
    )

    assert np.isclose(result, 4.5)


def test_estimate_rank_of_mrca_maximum_likelihood_with_common_ancestor7():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_maximum_likelihood(
        coincident_ranks, p_differentia_collision, hstrat.GeometricPrior(1.1)
    )

    assert np.isclose(result, 6)



import numpy as np

from hstrat._auxiliary_lib import unfurl_lineage_with_contiguous_ids


def test_unfurl_lineage():
    ancestor_ids = np.array([3, 2, 2, 3, 3, 1])
    leaf_id = 5
    expected_result = np.array([5, 1, 2])
    result = unfurl_lineage_with_contiguous_ids(ancestor_ids, leaf_id)
    np.testing.assert_array_equal(result, expected_result)


def test_unfurl_lineage_root():
    ancestor_ids = np.array([3, 2, 2, 3, 3, 1])
    leaf_id = 3
    expected_result = np.array([3])
    result = unfurl_lineage_with_contiguous_ids(ancestor_ids, leaf_id)
    np.testing.assert_array_equal(result, expected_result)


def test_unfurl_lineage_singleton():
    ancestor_ids = np.array([0])
    leaf_id = 0
    expected_result = np.array([0])
    result = unfurl_lineage_with_contiguous_ids(ancestor_ids, leaf_id)
    np.testing.assert_array_equal(result, expected_result)



import io
import warnings

import pandas as pd

from ._alifestd_find_chronological_inconsistency import (
    alifestd_find_chronological_inconsistency,
)
from ._alifestd_is_chronologically_sorted import (
    alifestd_is_chronologically_sorted,
)
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids


def _report_diagnosis(df: pd.DataFrame, taxon_id: int) -> None:
    warning_lines = [""]
    append = warning_lines.append

    append("chronological inconsitency found!")
    append("=================================")
    append(f"in phylogeny with {len(df)} rows")
    buf = io.StringIO()
    df.info(buf=buf, verbose=True)
    append(buf.getvalue())
    append(f"id values range from {df['id'].min()} to {df['id'].max()}")
    append(f"there are {df['id'].isna().sum()} nan id values")
    append(f"there are {df['id'].nunique()} unique id values")
    append(
        f"with {df['id'].is_monotonic_increasing=} "
        f"{df['id'].is_monotonic_decreasing=}",
    )
    append(
        f"and {alifestd_is_chronologically_sorted(df)=}, "
        f"and {alifestd_is_topologically_sorted(df)=}",
    )

    ot = df["origin_time"]
    append(f"origin_time values range from {ot.min()} to {ot.max()}")
    append(f"there are {ot.nunique()} unique origin_time values")
    append(f"there are {ot.isna().sum()} nan origin_time values")
    append(
        f"with {ot.is_monotonic_increasing=} {ot.is_monotonic_decreasing=}",
    )

    append("")
    append("")
    append(f"chronological inconsistency occurs at {taxon_id=}")
    (index,) = df[df["id"] == taxon_id].index
    loc = df.index.get_loc(index)
    append(f"for {taxon_id=}, position is {index=} {loc=}")

    origin_time = df.at[index, "origin_time"]
    append(f"for {taxon_id=}, {origin_time=}")
    taxon_origin_time = origin_time

    if "ancestor_id" in df.columns:
        ancestor_id = df.at[index, "ancestor_id"]
        append(f"for {taxon_id=}, {ancestor_id=}")

    ancestor_list = df.at[index, "ancestor_list"]
    ancestor_ids = alifestd_parse_ancestor_ids(ancestor_list)
    append(f"for {taxon_id=}, {ancestor_list=} and {ancestor_ids=}")

    for i, ancestor_id in enumerate(ancestor_ids):
        append("")
        (index,) = df[df["id"] == ancestor_id].index
        loc = df.index.get_loc(index)
        append(f"for {i}th {ancestor_id=}, position is {index=} {loc=}")

        origin_time = df.at[index, "origin_time"]
        prefix = f"for {i}th {ancestor_id=}, {origin_time=} so "
        ancestor_origin_time = origin_time
        time_delta = taxon_origin_time - ancestor_origin_time
        append(f"{prefix}{time_delta=}")
        if time_delta < 0:
            append(">" * len(prefix) + " ^^^^^^^^^ time delta is negative!")

    if "ancestor_id" in df.columns:
        ancestor_origin_time = df["ancestor_id"].map(
            df.set_index("id")["origin_time"],
        )
        discrepancy_mask = ancestor_origin_time > df["origin_time"]
        num_discrepancies = discrepancy_mask.sum()
        assert num_discrepancies, num_discrepancies
        append("")
        append(f"there are {num_discrepancies=} total")

    warnings.warn("\n".join(warning_lines))


def alifestd_is_chronologically_ordered(
    phylogeny_df: pd.DataFrame,
    diagnose: bool = True,
) -> bool:
    """Do any organisms have `origin_time`s preceding members of their
    `ancestor_list`?

    Input dataframe is not mutated by this operation.
    """

    res = alifestd_find_chronological_inconsistency(phylogeny_df)
    if diagnose and res is not None:
        _report_diagnosis(phylogeny_df, res)
    return res is None



import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ...juxtaposition import calc_rank_of_first_retained_disparity_between
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)


def does_have_any_common_ancestor(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    confidence_level: float = 0.95,
) -> typing.Optional[bool]:
    """Determine if common ancestry is evidenced with second.

    If insufficient common ranks between first and second are available to
    resolve any common ancestor, returns None.

    Note that stratum rention policies are strictly required to permanently
    retain the most ancient stratum.

    Parameters
    ----------
    confidence_level : float, optional
        The probability that we will correctly conclude no common ancestor
        is shared with second if, indeed, no common ancestor is actually
        shared. Default 0.95.

    See Also
    --------
    does_definitively_have_no_common_ancestor :
        Can we definitively conclude that first and second share no common
        ancestor?
    """
    if (
        calc_rank_of_earliest_detectable_mrca_between(
            first,
            second,
            confidence_level=confidence_level,
        )
        is None
    ):
        return None

    first_disparity = calc_rank_of_first_retained_disparity_between(
        first,
        second,
        confidence_level=confidence_level,
    )
    return True if first_disparity is None else first_disparity > 0



import pandas as pd

from hstrat._auxiliary_lib import alifestd_make_empty
import hstrat.phylogenetic_inference.tree._impl as impl


def test_append_genesis_organism_empty():
    df = alifestd_make_empty()
    df["origin_time"] = []

    expected_df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "origin_time": [0.0],
        }
    )

    df_ = df.copy()
    extended_df = impl.append_genesis_organism(df)

    assert df.equals(df_)
    assert extended_df.sort_index(axis=1).equals(
        expected_df.sort_index(axis=1)
    )


def test_append_genesis_organism_singleton():
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": ["[none]"],
            "origin_time": [1],
        }
    )
    expected_df = pd.DataFrame(
        {
            "id": [0, 1],
            "ancestor_list": ["[1]", "[none]"],
            "origin_time": [1, 0],
        }
    )

    df_ = df.copy()
    extended_df = impl.append_genesis_organism(df)

    assert df.equals(df_)
    assert extended_df.equals(expected_df)


def test_append_genesis_organism_one_root():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[1]", "[none]", "[1]", "[2]", "[2]"],
            "origin_time": [8, 1, 2, 1, 11],
            "name": ["A", "B", "C", "D", "E"],
        }
    )
    expected_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": ["[1]", "[5]", "[1]", "[2]", "[2]", "[none]"],
            "origin_time": [8, 1, 2, 1, 11, 0],
            "name": ["A", "B", "C", "D", "E", "genesis"],
        }
    )

    df_ = df.copy()
    extended_df = impl.append_genesis_organism(df)

    assert df.equals(df_)
    assert extended_df.equals(expected_df)


def test_append_genesis_organism_two_roots():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[1]", "[none]", "[none]", "[2]", "[2]"],
            "origin_time": [8, 1, 1, 2, 11],
            "name": ["A", "B", "C", "D", "E"],
        }
    )
    expected_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": ["[1]", "[5]", "[5]", "[2]", "[2]", "[none]"],
            "origin_time": [8, 1, 1, 2, 11, 0],
            "name": ["A", "B", "C", "D", "E", "genesis"],
        }
    )

    df_ = df.copy()
    extended_df = impl.append_genesis_organism(df)

    assert df.equals(df_)
    assert extended_df.equals(expected_df)


def test_append_genesis_organism_redundant():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": ["[1]", "[none]", "[1]", "[2]", "[2]"],
            "origin_time": [8, 0, 2, 1, 11],
            "name": ["A", "B", "C", "D", "E"],
        }
    )
    expected_df = df.copy()

    df_ = df.copy()
    extended_df = impl.append_genesis_organism(df)

    assert df.equals(df_)
    assert extended_df.equals(expected_df)



import itertools as it
import typing
import warnings

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ..pairwise import calc_rank_of_earliest_detectable_mrca_between


def calc_rank_of_earliest_detectable_mrca_among(
    population: typing.Iterable[HereditaryStratigraphicArtifact],
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """After what generation is common ancstry robustly detectable?

    Calculates the earliest possible rank a MRCA among the population
    could be reliably detected at.

    Even if a true MRCA of the population exists, if it occured earlier
    than the rank calculated here it could not be reliably detected with
    sufficient confidence after accounting for the possibility of spurious
    differentia collisions. (Although subsequent spurious differentia
    collisions after the true MRCA of first and second could lead to MRCA
    detection at such a rank.)

    Returns None if insufficient common ranks exist with population
    to ever conclude at the given confidence level the existance of any
    common ancestry among th epopulation (even if all strata at common
    ranks had equivalent differentiae).

    Also returns None if population is empty or singleton.

    Notes
    -----
    Currently implementaiton uses a naive O(n^2) approach. A more efficient
    implementation should be possible.
    """

    earliest_detectable_mrca_rank = -1
    any_combinations = False
    for (first, second) in it.combinations(population, 2):
        any_combinations = True
        assert (
            first.GetStratumDifferentiaBitWidth()
            == second.GetStratumDifferentiaBitWidth()
        )
        earliest_detectable_mrca_rank_between = (
            calc_rank_of_earliest_detectable_mrca_between(
                first, second, confidence_level
            )
        )
        if earliest_detectable_mrca_rank_between is None:
            return None
        else:
            earliest_detectable_mrca_rank = max(
                earliest_detectable_mrca_rank,
                earliest_detectable_mrca_rank_between,
            )

    if not any_combinations:
        warnings.warn(
            "Empty or singleton population. Unable to calculate rank of "
            "earliest detectable MRCA."
        )
        return None

    return earliest_detectable_mrca_rank



from typing import Any, Callable, Dict, List, Optional, Tuple

import lazy_loader

from ._GetAttrLaunderShim import GetAttrLaunderShim


def lazy_attach(
    module_name: str,
    *,
    submodules: Optional[List[str]],
    submod_attrs: Optional[Dict[str, List[str]]],
    should_launder: Callable = lambda _: True,
) -> Tuple[Callable[[str], Any], Callable[[], List[str]], List[str]]:
    """
    Attaches lazy loading to a package to reduce load times by deferring
    loading of submodules and their attributes until needed. This is used
    throughout the project in when a package imports all attributes from
    a subpackage (otherwise, use _lazy_attach_stub). Called in the
    __init__.py file of the package.

    Parameters
    ----------
    module_name : str
        The name of the package in which symbols are being imported. This
        should be `__name__` to properly set up lazy loading.
    submodules : list of str
        A list of submodules within the package that are to be loaded lazily.
    submod_attrs : dict
        A dictionary where keys are submodule names and values are lists
        of attribute names within each submodule to be loaded in the namespace.
    should_launder : callable, default `lambda x: True`
        A callable that takes a string attr and returns a boolean indicating
        whether the attr should be laundered. If True, __module__ on the
        attr will attempt to be set to `module_name`.

    Returns
    -------
    tuple
        A tuple containing the items needed to define `__getattr__`, `__dir__`,
        and `__all__` in the main module:

        - `__getattr__` : function
            Retrieves attributes lazily as needed.
        - `__dir__` : function
            Lists all available attributes, including lazily loaded ones.
        - `__all__` : list of str
            A list of all attributes for the namespace to be included in *
            imports and accessbile from other modules.

    Notes
    -----
    For more details on using `attach_lazy_loader`, refer to the
    Scientific Python project documentation:
    https://scientific-python.org/specs/spec-0001/

    This function leverages lazy loading to significantly reduce load
    times for large libraries by only loading submodules and attributes
    as they are accessed.

    See Also
    --------
    ._lazy_attach_stub : When there are no `__all__` imports, this function
        is preferred instead, as it only depends on a type stub.
    ._launder_impl_modules : Implementation details for laundering module
    names for attributes.
    """
    getattr__, dir__, all__ = lazy_loader.attach(
        module_name, submodules=submodules, submod_attrs=submod_attrs
    )
    wrapped_getattr = GetAttrLaunderShim(
        getattr__, should_launder, module_name
    )
    return wrapped_getattr, dir__, all__



from ._auxiliary_lib import get_hstrat_version

if __name__ == "__main__":
    print(f"hstrat v{get_hstrat_version()}")
    print()
    print("Available commands:")
    print("$ python3 -m hstrat.dataframe.surface_unpack_reconstruct")
    print()
    print(
        "For information on a command, " "invoke it with the --help flag.",
    )



import os

import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_is_topologically_sorted,
    alifestd_parse_ancestor_ids,
    alifestd_topological_sort,
    alifestd_validate,
    swap_rows_and_indices,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_topological_sort_empty(phylogeny_df):
    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df_ = phylogeny_df.copy()

    operand = phylogeny_df.iloc[-1:0, :]
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)
    # ensure no side effects
    assert phylogeny_df.equals(phylogeny_df_)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_topological_sort_singleton(phylogeny_df):
    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df_ = phylogeny_df.copy()

    operand = phylogeny_df.iloc[0:1, :]
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)
    # ensure no side effects
    assert phylogeny_df.equals(phylogeny_df_)


def test_alifestd_topological_sort_tworoots():

    phylo1 = pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv")
    phylo1.sort_values("id", ascending=True, inplace=True)
    phylo2 = pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv")
    phylo2.sort_values("id", ascending=True, inplace=True)

    operand = alifestd_aggregate_phylogenies(
        [
            phylo1.iloc[0:1, :],
            phylo2.iloc[0:1, :],
        ]
    )
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_topological_sort_twolineages(phylogeny_df):

    phylogeny_df.sort_values("id", ascending=True, inplace=True)
    phylogeny_df.reset_index(inplace=True)

    aggregated_df = alifestd_aggregate_phylogenies(
        [
            phylogeny_df.copy(),
            phylogeny_df.copy(),
        ]
    )

    lineage1 = aggregated_df[len(phylogeny_df) :]
    lineage2 = aggregated_df[: len(phylogeny_df)]

    operand = alifestd_aggregate_phylogenies(
        [
            lineage1.iloc[::-1, :],
            lineage2,
        ]
    )
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)

    operand = alifestd_aggregate_phylogenies(
        [
            lineage1.iloc[::-1, :],
            lineage2,
            lineage1.iloc[::-1, :],
        ]
    )
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)

    operand = alifestd_aggregate_phylogenies(
        [
            lineage1.iloc[::, :],
            lineage2,
            lineage1.iloc[::, :],
        ]
    )
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)


def test_alifestd_topologically_sort_twolineages_sexual():
    operand = pd.read_csv(
        f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
    )
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)

    operand = operand[::-1]
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_topological_sort(phylogeny_df):
    phylogeny_df.sort_values("id", ascending=False, inplace=True)
    phylogeny_df.set_index("id", drop=False, inplace=True)
    phylogeny_df_ = phylogeny_df.copy()

    operand = phylogeny_df
    res = alifestd_topological_sort(operand)
    assert alifestd_validate(res)
    assert alifestd_is_topologically_sorted(res)
    assert len(res) == len(operand)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)

    # reverse dataframe
    phylogeny_df = phylogeny_df.iloc[::-1]
    phylogeny_df_ = phylogeny_df.copy()
    assert alifestd_is_topologically_sorted(phylogeny_df)

    # one-by-one transpositions
    for idx, row in phylogeny_df.sample(min(10, len(phylogeny_df))).iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            operand = swap_rows_and_indices(phylogeny_df, idx, ancestor_id)
            res = alifestd_topological_sort(operand)
            assert alifestd_validate(res)
            assert alifestd_is_topologically_sorted(res)
            assert len(res) == len(operand)
            # check for side effects
            assert phylogeny_df.equals(phylogeny_df_)

    # cumulative transpositions in random order
    for idx, row in phylogeny_df.sample(min(10, len(phylogeny_df))).iterrows():
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            assert ancestor_id in phylogeny_df_.index
            phylogeny_df = swap_rows_and_indices(
                phylogeny_df, idx, ancestor_id
            )
            res = alifestd_topological_sort(operand)
            assert alifestd_validate(res)
            assert alifestd_is_topologically_sorted(res)
            assert len(res) == len(operand)



import numpy as np

from hstrat._auxiliary_lib import to_tril


def test_to_tril_square_matrix():
    # Test square matrix input
    matrix = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0], [7.0, 8.0, 9.0]])
    expected_output = [[0.0], [4.0, 0.0], [7.0, 8.0, 0.0]]
    assert to_tril(matrix) == expected_output


def test_to_tril_non_square_matrix():
    # Test non-square matrix input
    matrix = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    expected_output = [[0.0], [4.0, 0.0]]
    assert to_tril(matrix) == expected_output


def test_to_tril_square_matrix_with_zeros():
    # Test square matrix with zeros
    matrix = np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
    expected_output = [[0.0], [0.0, 0.0], [0.0, 0.0, 0.0]]
    assert to_tril(matrix) == expected_output



import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import alifestd_unfurl_lineage_asexual


def test_alifestd_unfurl_lineage_asexual():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [3, 3, 4, 4, 5, 5],
        }
    )
    leaf_id = 1
    expected_result = np.array([1, 3, 4, 5])

    for df in phylogeny_df, phylogeny_df.sample(frac=1):
        df_ = df.copy()
        result = alifestd_unfurl_lineage_asexual(df, leaf_id)
        assert np.array_equal(expected_result, result)
        assert df.equals(df_)


def test_alifestd_unfurl_lineage_asexual_singleton():
    phylogeny_df = pd.DataFrame(
        {
            "id": [7],
            "ancestor_id": [7],
        }
    )
    leaf_id = 7
    expected_result = np.array([7])

    df_ = phylogeny_df.copy()
    result = alifestd_unfurl_lineage_asexual(phylogeny_df, leaf_id)
    assert np.array_equal(expected_result, result)
    assert phylogeny_df.equals(df_)



from hstrat._auxiliary_lib import bytes_swap_bit_order


def test_bytes_swap_bit_order_single_byte():
    input_bytes = b"\x01"  # Binary: 00000001
    expected = b"\x80"  # Binary: 10000000
    assert bytes_swap_bit_order(input_bytes) == expected


def test_bytes_swap_bit_order_multiple_bytes():
    input_bytes = b"\x01\x02"  # Binary: 00000001 00000010
    expected = b"\x80\x40"  # Binary: 10000000 01000000
    assert bytes_swap_bit_order(input_bytes) == expected


def test_bytes_swap_bit_order_zero_byte():
    input_bytes = b"\x00"  # Binary: 00000000
    expected = b"\x00"  # Binary: 00000000
    assert bytes_swap_bit_order(input_bytes) == expected


def test_bytes_swap_bit_order_all_ones_byte():
    input_bytes = b"\xFF"  # Binary: 11111111
    expected = b"\xFF"  # Binary: 11111111
    assert bytes_swap_bit_order(input_bytes) == expected


def test_bytes_swap_bit_order_empty_bytes():
    input_bytes = b""  # Empty byte string
    expected = b""  # Empty byte string
    assert bytes_swap_bit_order(input_bytes) == expected



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase


class CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank:
    def __init__(
        self: "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
        policy_spec: typing.Optional[typing.Any] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
    ) -> int:
        """Calculate absolute MRCA uncertainty at pessimal rank."""
        pessimal_rank = policy.CalcMrcaUncertaintyAbsUpperBoundPessimalRank(
            first_num_strata_deposited,
            second_num_strata_deposited,
        )

        return policy.CalcMrcaUncertaintyAbsUpperBound(
            first_num_strata_deposited,
            second_num_strata_deposited,
            pessimal_rank,
        )



import copy
from typing import Dict

import pytest

from hstrat.serialization._impl import col_records_from_pop_records


@pytest.fixture()
def pop_records() -> Dict:
    return {
        "columns": [
            {"id": 1, "name": "column_1"},
            {"id": 2, "name": "column_2"},
            {"id": 3, "name": "column_3"},
        ],
        "policy": "some_policy",
        "policy_algo": "some_algo",
        "policy_spec": "some_spec",
        "differentia_bit_width": 16,
        "hstrat_version": 1.0,
    }


def test_col_records_from_pop_records(pop_records: Dict):
    col_records = list(col_records_from_pop_records(pop_records))
    assert len(col_records) == 3

    for col_record in col_records:
        assert isinstance(col_record, dict)
        assert "id" in col_record
        assert "name" in col_record
        assert col_record["policy"] == "some_policy"
        assert col_record["policy_algo"] == "some_algo"
        assert col_record["policy_spec"] == "some_spec"
        assert col_record["differentia_bit_width"] == 16
        assert col_record["hstrat_version"] == 1.0


def test_col_records_from_pop_records_mutate(pop_records: Dict):
    col_records = list(col_records_from_pop_records(pop_records, mutate=True))
    assert len(col_records) == 3

    for col_record in col_records:
        assert isinstance(col_record, dict)
        assert "id" in col_record
        assert "name" in col_record
        assert col_record["policy"] == "some_policy"
        assert col_record["policy_algo"] == "some_algo"
        assert col_record["policy_spec"] == "some_spec"
        assert col_record["differentia_bit_width"] == 16
        assert col_record["hstrat_version"] == 1.0

    assert pop_records["columns"] == col_records


def test_col_records_from_pop_records_empty():
    pop_records = {"columns": []}
    col_records = list(col_records_from_pop_records(pop_records))
    assert len(col_records) == 0


def test_col_records_from_pop_records_no_mutate(pop_records: Dict):
    pop_records_copy = copy.deepcopy(pop_records)
    col_records = list(
        col_records_from_pop_records(pop_records_copy, mutate=False)
    )
    assert len(col_records) == 3

    for col_record in col_records:
        assert isinstance(col_record, dict)
        assert "id" in col_record
        assert "name" in col_record
        assert col_record["policy"] == "some_policy"
        assert col_record["policy_algo"] == "some_algo"
        assert col_record["policy_spec"] == "some_spec"
        assert col_record["differentia_bit_width"] == 16
        assert col_record["hstrat_version"] == 1.0

    assert pop_records == pop_records_copy



import random
import typing

from ._unzip import unzip


def coshuffled(*args) -> typing.Tuple:
    """Apply a consistent shuffle to multiple sequences.

    Parameters
    ----------
    *args : sequence of array-like objects
        Sequence of array-like objects that need to be shuffled together.

    Returns
    -------
    tuple
        A tuple with shuffled copies of `args` array-like objects.

    Notes
    -----
    This function shuffles multiple sequences of data together using the `zip` and `shuffle` functions from the built-in
    `random` module.

    Examples
    --------
    >>> x = [1, 2, 3]
    >>> y = [4, 5, 6]
    >>> z = [7, 8, 9]
    >>> coshuffled(x, y, z)
    ([2, 3, 1], [5, 6, 4], [8, 9, 7])
    """
    # if len(args) == 1:

    zipped = [*zip(*args)]
    random.shuffle(zipped)
    return tuple(map(list, unzip(zipped))) or tuple([] for __ in args)



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase


class CalcNumStrataRetainedUpperBoundWorstCase:
    def __init__(
        self: "CalcNumStrataRetainedUpperBoundWorstCase",
        policy_spec: typing.Optional[typing.Any] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcNumStrataRetainedUpperBoundWorstCase",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedUpperBoundWorstCase",
        policy: typing.Optional[PolicyCouplerBase],
        num_strata_deposited: int,
    ) -> int:
        """At most, how many strata are retained after n deposited? Inclusive."""
        return num_strata_deposited



import typing

import numpy as np
import pandas as pd

from ._alifestd_has_compact_ids import alifestd_has_compact_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._jit import jit
from ._jit_TypingError import jit_TypingError


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency_contiguous(
    ancestor_ids: np.ndarray, origin_times: np.ndarray
) -> typing.Optional[int]:
    for id_, ancestor_id in enumerate(ancestor_ids):
        # self comparison ok for geneses
        if origin_times[ancestor_id] > origin_times[id_]:
            return id_
    return None


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency_compact(
    ids: np.ndarray, ancestor_ids: np.ndarray, origin_times: np.ndarray
) -> typing.Optional[int]:
    origin_time_lookup = np.empty(len(ids))
    for id_, origin_time in zip(ids, origin_times):
        origin_time_lookup[id_] = origin_time

    for id_, ancestor_id in zip(ids, ancestor_ids):
        # self comparison ok for geneses
        if origin_time_lookup[ancestor_id] > origin_time_lookup[id_]:
            return id_
    return None


@jit(nopython=True)
def _alifestd_find_chronological_inconsistency_arbitrary(
    ids: np.ndarray, ancestor_ids: np.ndarray, origin_times: np.ndarray
) -> typing.Optional[int]:
    origin_time_lookup = dict()
    for id_, origin_time in zip(ids, origin_times):
        origin_time_lookup[id_] = origin_time

    for id_, ancestor_id in zip(ids, ancestor_ids):
        # self comparison ok for geneses
        if origin_time_lookup[ancestor_id] > origin_time_lookup[id_]:
            return id_
    return None


def _alifestd_find_chronological_inconsistency_asexual(
    phylogeny_df: pd.DataFrame, unwrap: typing.Callable
) -> typing.Optional[int]:
    if alifestd_has_contiguous_ids(phylogeny_df):
        call = unwrap(_alifestd_find_chronological_inconsistency_contiguous)
        return call(
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["origin_time"].to_numpy(),
        )
    elif alifestd_has_compact_ids(phylogeny_df):
        call = unwrap(_alifestd_find_chronological_inconsistency_compact)
        return call(
            phylogeny_df["id"].to_numpy(),
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["origin_time"].to_numpy(),
        )
    else:
        call = unwrap(_alifestd_find_chronological_inconsistency_arbitrary)
        return call(
            phylogeny_df["id"].to_numpy(),
            phylogeny_df["ancestor_id"].to_numpy(),
            phylogeny_df["origin_time"].to_numpy(),
        )


def alifestd_find_chronological_inconsistency(
    phylogeny_df: pd.DataFrame,
) -> typing.Optional[int]:
    """Return the id of a taxon with origin time preceding its parent's, if
    any are present."""

    if "ancestor_id" in phylogeny_df.columns:
        for unwrap in lambda x: x, lambda x: x.__wrapped__:
            try:
                return _alifestd_find_chronological_inconsistency_asexual(
                    phylogeny_df, unwrap
                )
            except jit_TypingError:  # pragma: no cover
                # this does get exercised in tests, but not in coverage
                # because jit is disabled for better visibility of jit code
                continue
            except ValueError:
                continue  # if origin_times are sequences i.e., tuples

    phylogeny_df = phylogeny_df.set_index("id", drop=False)

    origin_time_loc = phylogeny_df.columns.get_loc("origin_time")
    for _pos, (_idx, row) in enumerate(phylogeny_df.iterrows()):
        for ancestor_id in alifestd_parse_ancestor_ids(row["ancestor_list"]):
            ancestor_loc = phylogeny_df.index.get_loc(ancestor_id)
            if (
                phylogeny_df.iat[ancestor_loc, origin_time_loc]
                > row["origin_time"]
            ):
                return _idx

    return None



# adapted from https://cs.stackexchange.com/a/65744
def render_to_numeral_system(n: int, alphabet: str) -> str:
    """Convert an integer to its digit representation in a custom base.

    Parameters
    ----------
    n : int
        The non-negative integer to be converted to a custom base
        representation.
    alphabet : str
        A string of unique characters representing the digits in the custom
        base.

        The i-th character in the alphabet corresponds to the i-th digit in the
        base.

    Returns
    -------
    str
        A string of characters representing the custom base representation of the given integer.

    Raises
    ------
    AssertionError
        If the integer is negative or if the alphabet contains repeated
        characters.

    Examples
    --------
    >>> render_to_numeral_system(163, '0123456789abcdef')
    'a3'
    >>> render_to_numeral_system(42, '01234567')
    '52'
    >>> render_to_numeral_system(7, '01')
    '111'
    """
    assert n >= 0
    assert len(alphabet) == len(set(alphabet))
    if n == 0:
        return alphabet[0]
    b = len(alphabet)
    reverse_digits = []
    while n > 0:
        reverse_digits.append(alphabet[n % b])
        n = n // b
    return "".join(reversed(reverse_digits))



import typing

from iterpop import iterpop as ip
import pandas as pd

from . import trie_postprocess
from ..._auxiliary_lib import (
    HereditaryStratigraphicArtifact,
    alifestd_make_empty,
)
from ..priors._detail import PriorBase
from ._build_tree_trie_ensemble import build_tree_trie_ensemble
from .trie_postprocess._detail import TriePostprocessorBase


def build_tree_trie(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    force_common_ancestry: bool = False,
    progress_wrap: typing.Callable = lambda x: x,
    seed: typing.Optional[int] = 1,
    bias_adjustment: typing.Union[
        typing.Literal["sample_ancestral_rollbacks"],
        PriorBase,
        TriePostprocessorBase,
        None,
    ] = None,
) -> pd.DataFrame:
    """Estimate the phylogenetic history among hereditary stratigraphic
    artifacts by building a trie (a.k.a. prefix tree) of their differentiae
    records.

    The `build_tree_searchtable` function should be preferred, as it applies
    an equivalent, but more efficient, algorithm.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, default identity function
        Pass tqdm or equivalent to display progress bars.
    seed : int, default 1
        Controls tiebreaking decisions in the algorithm.

        Pass an int for reproducible output across multiple function calls. The
        default value, 1, ensures reproducible output. Pass None to use global
        RNG context.
    bias_adjustment : "sample_ancestral_rollbacks", PriorBase, or
    TriePostProcessorBase, optional
        How should bias toward overestimation of relatedness due to differentia
        collisions be corrected for?

        If "sample_ancestral_rollbacks", the trie topology will be adjusted as
        if the expected number of collisions had occurred. Targets for
        "unzipping" to reverse the effect of a speculated collision are
        chosen randomly from within the tree. See
        `SampleAncestralRollbacksTriePostprocessor` for details.

        If a prior functor is passed, the origin time for each trie node will
        be calculated as the expected origin time over the distribution of
        possible differentia collisions. Correction recursively takes into
        account the possibility of multiple collisions. See
        `hstrat.phylogenetic_inference.priors` for available prior
        distributions. A custom prior distribution may also be supplied. See
        `AssignOriginTimeExpectedValueTriePostprocessor` for details.

        If a prior functor is passed, correction for guaranteed-spurious
        collision between most-recent strata will also be performed. See
        `PeelBackConjoinedLeavesTriePostprocessor` for details.

        If None, no correction will be performed. The origin time for each trie
        node will be assigned using a naive strategy, calculated as the average
        of the node's rank and the minimum rank among its children. See
        `AssignOriginTimeNaiveTriePostprocessor` for details.

        If you want to use multiple postprocessors on the same tree, use the
        `build_tree_trie_ensemble` function directly, which returns a list of
        DataFrames resulting from the different postprocessors.

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.

    Notes
    -----
    Unifurcations in the reconstructed tree are collapsed.

    However, polytomies are not resolved. In addition to any true polytomies,
    ancestry sequences that cannot be resolved due to missing information
    appear as polytomies in the generated reconstruction. Therefore, polytomies
    are generally overrepresented in reconstructions, especially when low
    hereditary stratigraphic resolution is available. If overestimation of
    polytomies is problematic, external tools can be used to decompose
    polytomies into arbitrarily-arranged bifurcations.

    See Also
    --------
    build_tree_searchtable :
        Implementation using more efficient trie-based reconstruction algorithm.
    build_tree_trie_ensemble :
        Implementation function delegated to after postprocessors are determined.
        For multiple postprocessors, use it directly.
    """

    # for simplicity, return early for this special case
    if len(population) == 0:
        return alifestd_make_empty()

    if bias_adjustment is None:
        trie_postprocessor = (
            trie_postprocess.AssignOriginTimeNaiveTriePostprocessor()
        )
    elif bias_adjustment == "sample_ancestral_rollbacks":

        trie_postprocessor = trie_postprocess.CompoundTriePostprocessor(
            postprocessors=[
                trie_postprocess.SampleAncestralRollbacksTriePostprocessor(
                    seed=1,
                ),
                trie_postprocess.AssignOriginTimeNaiveTriePostprocessor(),
            ],
        )
    elif isinstance(bias_adjustment, PriorBase):
        trie_postprocessor = trie_postprocess.CompoundTriePostprocessor(
            postprocessors=[
                trie_postprocess.PeelBackConjoinedLeavesTriePostprocessor(),
                trie_postprocess.AssignOriginTimeExpectedValueTriePostprocessor(
                    prior=bias_adjustment,
                ),
            ],
        )
    elif isinstance(bias_adjustment, TriePostprocessorBase):
        trie_postprocessor = bias_adjustment
    else:
        raise TypeError(f"Provided {bias_adjustment=} has unrecognized type")

    return ip.popsingleton(
        build_tree_trie_ensemble(
            population=population,
            taxon_labels=taxon_labels,
            force_common_ancestry=force_common_ancestry,
            progress_wrap=progress_wrap,
            seed=seed,
            trie_postprocessors=[trie_postprocessor],
        ),
    )



import functools
import itertools as it
import os
import random

import alifedata_phyloinformatics_convert as apc
import dendropy as dp
import pandas as pd
import pytest
from tqdm import tqdm

from hstrat import hstrat

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "always_store_rank_in_stratum",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "num_predeposits",
    [
        0,
        1,
        10,
        100,
    ],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.stochastic_algo.Policy(retention_probability=0.05),
        hstrat.depth_proportional_resolution_algo.Policy(
            depth_proportional_resolution=10
        ),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "set_stem_length",
    [
        lambda node: 1,
        lambda node: 1 + random.randrange(10),
    ],
)
@pytest.mark.parametrize(
    "tree",
    [
        dp.Tree(),
        dp.Tree.get(
            path=f"{assets_path}/grandchild_and_aunt.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandchild_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(path=f"{assets_path}/grandchild.newick", schema="newick"),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets_and_aunt.newick",
            schema="newick",
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtriplets.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtwins_and_aunt.newick", schema="newick"
        ),
        dp.Tree.get(
            path=f"{assets_path}/grandtwins_and_auntuncle.newick",
            schema="newick",
        ),
        dp.Tree.get(path=f"{assets_path}/grandtwins.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/justroot.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/triplets.newick", schema="newick"),
        dp.Tree.get(path=f"{assets_path}/twins.newick", schema="newick"),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        ),
        apc.alife_dataframe_to_dendropy_tree(
            pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
        ),
    ],
)
def test_descend_template_phylogeny(
    always_store_rank_in_stratum,
    num_predeposits,
    retention_policy,
    set_stem_length,
    tree,
):
    # setup tree
    for node in tree:
        node.edge_length = set_stem_length(node)

    tree.seed_node.edge_length = num_predeposits

    for loc, node in enumerate(tree.leaf_node_iter()):
        node.taxon = tree.taxon_namespace.new_taxon(label=str(loc))

    tree.update_bipartitions(
        suppress_unifurcations=False,
        collapse_unrooted_basal_bifurcation=False,
    )

    # setup seed column
    seed_column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        always_store_rank_in_stratum=always_store_rank_in_stratum,
    )
    seed_column.DepositStrata(num_stratum_depositions=num_predeposits)

    extant_population = hstrat.descend_template_phylogeny(
        ascending_lineage_iterables=(
            tip_node.ancestor_iter(
                inclusive=True,
            )
            for tip_node in tree.leaf_node_iter()
        ),
        descending_tree_iterable=tree.levelorder_node_iter(),
        get_parent=lambda node: node.parent_node,
        get_stem_length=lambda node: node.edge_length,
        seed_column=seed_column,
        progress_wrap=functools.partial(tqdm, disable=True),
    )

    num_tips = len(tree)
    assert num_tips == len(extant_population)

    tip_depths = [
        int(tip_node.distance_from_root())
        for tip_node in tree.leaf_node_iter()
    ]
    assert tip_depths == [
        column.GetNumStrataDeposited() - 1 for column in extant_population
    ]

    sampled_product = it.permutations(
        random.sample(
            [*zip(extant_population, tree.leaf_node_iter())],
            min(10, len(extant_population)),
        ),
        2,
    )
    spliced_product = it.permutations(
        it.islice(zip(extant_population, tree.leaf_node_iter()), 10),
        2,
    )

    for (c1, n1), (c2, n2) in it.chain(sampled_product, spliced_product):
        lb, ub = hstrat.calc_rank_of_mrca_bounds_between(
            c1, c2, prior="arbitrary"
        )
        mrca = tree.mrca(
            taxa=[n1.taxon, n2.taxon],
            is_bipartitions_updated=True,
        )
        # patch for dendropy bug where internal unifurcations are not accounted
        # for in mrca detection
        # see https://github.com/jeetsukumaran/DendroPy/pull/148
        while mrca.num_child_nodes() == 1:
            (mrca,) = mrca.child_nodes()

        assert lb <= mrca.distance_from_root() < ub



import anytree

from hstrat._auxiliary_lib import anytree_cardinality


def test_empty_tree():
    # Create an empty tree
    root = anytree.Node("root")
    # Assert that the length of the tree is 1 (just the root node)
    assert anytree_cardinality(root) == 1


def test_single_node_tree():
    # Create a tree with just one node
    root = anytree.Node("root")
    # Assert that the length of the tree is 1
    assert anytree_cardinality(root) == 1


def test_two_node_tree():
    # Create a tree with two nodes
    root = anytree.Node("root")
    anytree.Node("child1", parent=root)
    # Assert that the length of the tree is 2 (root and child1)
    assert anytree_cardinality(root) == 2


def test_multi_level_tree():
    # Create a multi-level tree
    root = anytree.Node("root")
    child1 = anytree.Node("child1", parent=root)
    child2 = anytree.Node("child2", parent=root)
    anytree.Node("grandchild1", parent=child1)
    anytree.Node("grandchild2", parent=child1)
    anytree.Node("grandchild3", parent=child2)
    # Assert that the length of the tree is 6 (root, child1, child2, grandchild1, grandchild2, and grandchild3)
    assert anytree_cardinality(root) == 6



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcNumStrataRetainedExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcNumStrataRetainedExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcNumStrataRetainedExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedExact",
        policy: typing.Optional[PolicyCouplerBase],
        num_strata_deposited: int,
    ) -> int:
        """Exactly how many strata are retained after n deposited?"""
        return num_strata_deposited



from ._count_trailing_zeros import count_trailing_zeros


def count_trailing_ones(n: int) -> int:
    """Count the number of trailing ones in the binary representation of the
    absolute value of an integer.

    Trailing ones are contiguous set bits before the first unset bit at the least significant (rightmost) position. For example, the number 9 (1001 in binary) has 1 trailing one. Returns 0 trailing bits for 0.

    Examples
    --------
    >>> count_trailing_ones(9)
    1

    >>> count_trailing_ones(0)
    0
    """
    return count_trailing_zeros(~abs(n))



import typing

from ._calc_common_ratio import calc_common_ratio


def iter_target_recencies(
    degree: int,
    num_strata_deposited: int,
) -> typing.Iterator[int]:
    """Yield strata recencies for each exponentially-spaced coverage target
    `pow` in ascending order."""
    # target recencies are a geometric sequence
    common_ratio = calc_common_ratio(degree, num_strata_deposited)
    # don't iterate over 0th pow, this is just the most recent rank
    # i.e., recency == 1
    for pow_ in range(1, degree + 1):
        yield common_ratio**pow_



import random
from unittest.mock import MagicMock

import contexttimer as ctt
import numpy as np
import pytest

from hstrat import hstrat
from hstrat.juxtaposition._impl_column import (
    calc_rank_of_first_retained_disparity_between_generic,
)
from hstrat.juxtaposition._impl_specimen import (
    calc_rank_of_first_retained_disparity_between,
)


@pytest.fixture
def specimens_mock_simple():
    # Create example HereditaryStratigraphicAssemblageSpecimen instances
    # with known values for testing.
    first = MagicMock()
    second = MagicMock()
    first.GetStratumDifferentiaBitWidth.return_value = 8
    second.GetStratumDifferentiaBitWidth.return_value = 8
    first.GetDifferentiaVals.return_value = np.array(
        [5, 6, 7, 8, 9], dtype=np.uint32
    )
    second.GetDifferentiaVals.return_value = np.array(
        [5, 6, 7, 8, 9], dtype=np.uint32
    )
    first.GetNumStrataRetained.return_value = 5
    second.GetNumStrataRetained.return_value = 5
    first.GetNumStrataDeposited.return_value = 5
    second.GetNumStrataDeposited.return_value = 5
    first.GetRankIndex.return_value = np.array(
        [0, 1, 2, 3, 4], dtype=np.uint64
    )
    second.GetRankIndex.return_value = np.array(
        [0, 1, 2, 3, 4], dtype=np.uint64
    )
    return first, second


def test_calc_rank_of_first_retained_disparity_between_mock_simple(
    specimens_mock_simple,
):
    first, second = specimens_mock_simple

    # Test case where there is no disparity.
    assert (
        calc_rank_of_first_retained_disparity_between(
            first, second, confidence_level=0.49
        )
        is None
    )

    # Test case where there is disparity.
    second.GetDifferentiaVals.return_value = np.array(
        [5, 6, 3, 8, 9], dtype=np.uint32
    )
    assert (
        calc_rank_of_first_retained_disparity_between(
            first, second, confidence_level=0.49
        )
        == 2
    )


@pytest.fixture
def specimens_mock_complex():
    # Create example HereditaryStratigraphicAssemblageSpecimen instances
    # with known values for testing.
    first = MagicMock()
    second = MagicMock()
    first.GetStratumDifferentiaBitWidth.return_value = 8
    second.GetStratumDifferentiaBitWidth.return_value = 8
    first.GetDifferentiaVals.return_value = np.array(
        [5, 6, 7, 8, 9], dtype=np.uint32
    )
    second.GetDifferentiaVals.return_value = np.array(
        [5, 42, 9, 202], dtype=np.uint32
    )
    first.GetNumStrataRetained.return_value = 5
    second.GetNumStrataRetained.return_value = 4
    first.GetNumStrataDeposited.return_value = 5
    second.GetNumStrataDeposited.return_value = 6
    first.GetRankIndex.return_value = np.array(
        [0, 10, 20, 30, 40], dtype=np.uint64
    )
    second.GetRankIndex.return_value = np.array(
        [0, 20, 40, 50], dtype=np.uint64
    )
    return first, second


def test_calc_rank_of_first_retained_disparity_between_mock_complex(
    specimens_mock_complex,
):
    first, second = specimens_mock_complex

    assert (
        calc_rank_of_first_retained_disparity_between(
            first, second, confidence_level=0.49
        )
        == 20
    )

    assert (
        calc_rank_of_first_retained_disparity_between(
            first, second, confidence_level=0.999
        )
        == 0
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=7),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2,
        ),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64, 65],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.49, 0.8, 0.95, 0.9999],
)
@pytest.mark.parametrize(
    "uneven_branches",
    [True, False],
)
@pytest.mark.parametrize(
    "other_pop_member",
    [True, False],
)
def test_compare_to_generic_column_impl(
    retention_policy,
    differentia_width,
    confidence_level,
    uneven_branches,
    other_pop_member,
):
    common_ancestors = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
    ]
    for __ in range(37):
        common_ancestors.append(common_ancestors[-1].CloneDescendant())

    for _rep in range(250):
        _ = _rep
        num_total = random.randrange(0, 37)

        num_together = random.randrange(num_total + 1)
        assert 0 <= num_together <= num_total
        num_alone = num_total - num_together

        left_alone = num_alone
        right_alone = num_alone

        common_ancestor = common_ancestors[num_together]
        left = common_ancestor.Clone()
        right = common_ancestor.Clone()

        left.DepositStrata(left_alone)
        right.DepositStrata(
            right_alone + uneven_branches * random.randrange(0, 37)
        )

        pop = [left, right]
        if other_pop_member:
            pop.append(
                random.choice(common_ancestors).CloneNthDescendant(
                    random.randrange(0, 63)
                )
            )
        specimen1, specimen2, *__ = map(hstrat.col_to_specimen, pop)

        assert (
            specimen1.GetNumStrataDeposited() == left.GetNumStrataDeposited()
        )
        assert specimen1.GetNumStrataRetained() == left.GetNumStrataRetained()
        assert (
            specimen2.GetNumStrataDeposited() == right.GetNumStrataDeposited()
        )
        assert specimen2.GetNumStrataRetained() == right.GetNumStrataRetained()

        assert calc_rank_of_first_retained_disparity_between(
            specimen1, specimen2, confidence_level=confidence_level
        ) == calc_rank_of_first_retained_disparity_between_generic(
            left, right, confidence_level=confidence_level
        )


def test_benchmark():
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=8,
        stratum_retention_policy=hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10
        ),
    ).CloneNthDescendant(100000)

    tip1_c = common_ancestor.CloneNthDescendant(100000)
    tip2_c = common_ancestor.CloneNthDescendant(90000)

    tip1_s = hstrat.col_to_specimen(tip1_c)
    tip2_s = hstrat.col_to_specimen(tip2_c)

    # pre-compile jit outside of timing
    calc_rank_of_first_retained_disparity_between(
        tip1_s,
        tip2_s,
        confidence_level=0.9999,
    )

    with ctt.Timer(factor=1000) as t_algo_dstruct:
        for __ in range(10000):
            calc_rank_of_first_retained_disparity_between(
                tip1_s,
                tip2_s,
                confidence_level=0.9999,
            )

    # note higher factor and lower repcount
    with ctt.Timer(factor=10000) as t_dstruct:
        for __ in range(1000):
            calc_rank_of_first_retained_disparity_between_generic(
                tip1_s,
                tip2_s,
                confidence_level=0.9999,
            )

    with ctt.Timer(factor=1000) as t_vanilla:
        for __ in range(10000):
            calc_rank_of_first_retained_disparity_between_generic(
                tip1_c,
                tip2_c,
                confidence_level=0.9999,
            )

    print(f"t_algo_dstruct={t_algo_dstruct}")
    print(f"t_dstruct={t_dstruct} ")
    print(f"t_vanilla={t_vanilla}")



import contexttimer as ctt
import interval_search as inch
import numpy as np

from hstrat._auxiliary_lib import curried_binary_search_jit, jit


def test_curried_binary_search_jit_singleton():
    @jit(nopython=True)
    def predicate(i: int) -> bool:
        return True

    assert curried_binary_search_jit(predicate)(10, 10) == 10


def test_binary_search_jit():
    @jit(nopython=True)
    def predicate(i: int) -> bool:
        return i >= 5

    assert curried_binary_search_jit(predicate)(0, 100) == 5


def test_curried_binary_search_jit_numpy():
    data = np.arange(100)

    @jit(nopython=True)
    def predicate(i: int) -> bool:
        return data[i] > data[::-1][i]

    assert curried_binary_search_jit(predicate)(0, 100) == inch.binary_search(
        predicate, 0, 100
    )


def test_fruitless_curried_binary_search_jit():
    @jit(nopython=True)
    def predicate_false(i: int) -> bool:
        return False

    assert curried_binary_search_jit(predicate_false)(0, 100) is None


def test_empty_curried_binary_search_jit():
    @jit(nopython=True)
    def predicate_false(i: int) -> bool:
        return False

    @jit(nopython=True)
    def predicate_true(i: int) -> bool:
        return False

    assert curried_binary_search_jit(predicate_false)(100, 99) is None
    assert curried_binary_search_jit(predicate_false)(100, 0) is None
    assert curried_binary_search_jit(predicate_true)(100, 99) is None
    assert curried_binary_search_jit(predicate_true)(100, 0) is None


def test_benchmark():
    @jit(nopython=True)
    def predicate(i: int) -> bool:
        return i <= 5

    do_search = curried_binary_search_jit(predicate)
    do_search(0, 2**31)

    with ctt.Timer(factor=1000) as t_jit:
        for __ in range(10000):
            do_search(0, 2**31)

    with ctt.Timer(factor=1000) as t_vanilla:
        for __ in range(10000):
            inch.binary_search(lambda x: x <= 5, 0, 2**31)

    with ctt.Timer(factor=1000) as t_vanilla_jitpred:
        for __ in range(10000):
            inch.binary_search(predicate, 0, 2**31)

    print(f"t_jit={t_jit}")
    print(f"t_vanilla={t_vanilla}")
    print(f"t_vanilla_jitpred={t_vanilla_jitpred}")



from ._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce,
)
from ._CalcMrcaUncertaintyAbsUpperBoundWorstCase import (
    CalcMrcaUncertaintyAbsUpperBoundWorstCase,
)
from ._CalcMrcaUncertaintyRelExactFromAbs import (
    CalcMrcaUncertaintyRelExactFromAbs,
)
from ._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce,
)
from ._CalcMrcaUncertaintyRelUpperBoundWorstCase import (
    CalcMrcaUncertaintyRelUpperBoundWorstCase,
)
from ._CalcNumStrataRetainedUpperBoundWorstCase import (
    CalcNumStrataRetainedUpperBoundWorstCase,
)
from ._GenDropRanksFromPredKeepRank import GenDropRanksFromPredKeepRank

__all__ = [
    "CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyAbsUpperBoundPessimalRankBruteForce",
    "CalcMrcaUncertaintyRelExactFromAbs",
    "CalcMrcaUncertaintyRelUpperBoundAtPessimalRank",
    "CalcMrcaUncertaintyRelUpperBoundPessimalRankBruteForce",
    "CalcMrcaUncertaintyAbsUpperBoundWorstCase",
    "CalcMrcaUncertaintyRelUpperBoundWorstCase",
    "CalcNumStrataRetainedUpperBoundWorstCase",
    "GenDropRanksFromPredKeepRank",
]



import typing

from ...._auxiliary_lib import (
    AnyTreeFastLeafIter,
    anytree_iterative_deepcopy,
    anytree_peel_sibling_to_cousin,
)
from .._impl import TrieInnerNode
from ._detail import TriePostprocessorBase


class PeelBackConjoinedLeavesTriePostprocessor(TriePostprocessorBase):
    """Functor to separate any TrieLeafNode instances that are direct siblings.

    Corrects for guaranteed-spurious differentia collisions among most-recent
    strata.
    """

    def __call__(
        self: "PeelBackConjoinedLeavesTriePostprocessor",
        trie: TrieInnerNode,
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> TrieInnerNode:
        """Peel apart any `TrieLeafNode` nodes that are direct siblings.

        Without reproduction dynamics that allow columns to be cloned without
        stratum deposit, two hereditary stratigraphic columns can only share
        their most-recent strata by collision.

        Clones the sibling leaves' parent node and attaches it to the
        the siblings' grandparent node. One sibling node is then grafted away
        from its original parent and attached onto the newly cloned parent node
        as a child. The original parent node is left in place with any
        remaining children. This process is repeated until no `TrieLeafNode`'s
        remain as direct siblings.

        Parameters:
        ----------
        trie : TrieInnerNode
            The root node of the trie to be unzipped.
        p_differentia_collision : float
            The multiplicative inverse of the number of possible
            differentia.

            This fraction of possible rollbacks are performed.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode
            The postprocessed trie.
        """
        if not mutate:
            trie = anytree_iterative_deepcopy(
                trie, progress_wrap=progress_wrap
            )

        for leaf in progress_wrap(AnyTreeFastLeafIter(trie)):
            if sum(1 for __ in leaf.parent.outer_children) > 1:
                anytree_peel_sibling_to_cousin(leaf)

        return trie



"""Data structures to annotate genomes with."""

from _HereditaryStratigraphicColumn import HereditaryStratigraphicColumn
from _HereditaryStratigraphicColumnBundle import (
    HereditaryStratigraphicColumnBundle,
)
from _HereditaryStratum import HereditaryStratum
from stratum_ordered_stores import (
    HereditaryStratumOrderedStoreDict,
    HereditaryStratumOrderedStoreList,
    HereditaryStratumOrderedStoreTree,
    provided_stratum_ordered_stores,
)

from . import stratum_ordered_stores

__all__ = [
    "stratum_ordered_stores",
    "HereditaryStratum",
    "HereditaryStratigraphicColumn",
    "HereditaryStratigraphicColumnBundle",
    "HereditaryStratumOrderedStoreDict",
    "HereditaryStratumOrderedStoreList",
    "HereditaryStratumOrderedStoreTree",
    "provided_stratum_ordered_stores",
]



from functools import lru_cache
import warnings


@lru_cache(maxsize=None)
def warn_once(msg: str) -> None:
    """Wraps `warnings.warn`, dropping repeat calls with same `msg`."""
    warnings.warn(msg)



import numpy as np

from hstrat.phylogenetic_inference.tree._impl._TrieInnerNode import (
    TrieInnerNode,
)
from hstrat.phylogenetic_inference.tree._impl._build_trie_from_artifacts import (
    MatrixColumn,
)


def is_matrix_equal_trie(m: np.ndarray, root: TrieInnerNode) -> bool:
    # stack based approach to go through each root
    # the matrix approach should operate exactly the same as the normal one
    # therefore, the order in which children are added should be the same
    roots_to_analyze = [(root, 1)]
    while roots_to_analyze:
        r, i = roots_to_analyze.pop()
        if not (
            r.rank == m[i][MatrixColumn.RANK.value]
            and r.differentia == m[i][MatrixColumn.DIFFERENTIA.value]
        ):
            return False
        inner = [*r.inner_children]
        inner_m = []
        child_index = m[i][MatrixColumn.FIRST_CHILD_ID.value]
        while child_index:
            if m[child_index][MatrixColumn.IS_LEAF_NODE.value] == 0:
                inner_m.append(child_index)
            child_index = m[child_index][MatrixColumn.NEXT_SIBLING_ID.value]
        for a, b in zip(inner, inner_m):
            if not int(a.differentia) == m[b][MatrixColumn.DIFFERENTIA.value]:
                return False

        outer = [*r.outer_children]
        outer_m = []
        child_index = m[i][MatrixColumn.FIRST_CHILD_ID.value]
        while child_index:
            if m[child_index][MatrixColumn.IS_LEAF_NODE.value] == 1:
                outer_m.append(child_index)
            child_index = m[child_index][MatrixColumn.NEXT_SIBLING_ID.value]
        for a, b in zip(outer, outer_m):
            if (
                not int(a.taxon_label)
                == m[b][MatrixColumn.TAXON_LABEL_ID.value]
            ):
                return False

        roots_to_analyze.extend([*zip(inner, inner_m)])

    return True



import unittest

from hstrat._auxiliary_lib import is_nonincreasing


class TestIsNonincreasing(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert is_nonincreasing([])

    def test_singleton(self):
        assert is_nonincreasing(["a"])
        assert is_nonincreasing([0])
        assert is_nonincreasing([1])

    def test_nondecreasing(self):
        assert is_nonincreasing(
            reversed(
                [
                    *range(10),
                ]
            )
        )
        assert is_nonincreasing(
            reversed(
                [
                    0,
                    *range(10),
                ]
            )
        )
        assert is_nonincreasing(
            reversed(
                [
                    0,
                    0,
                    *range(10),
                    *range(9, 18),
                ]
            )
        )

    def test_decreasing(self):
        assert not is_nonincreasing(
            [
                -1,
                0,
            ]
        )
        assert not is_nonincreasing(
            reversed(
                [
                    *range(10),
                    *range(2),
                ]
            )
        )


if __name__ == "__main__":
    unittest.main()



import itertools as it
import warnings

import pandas as pd

from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_splay_polytomies(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Use a simple splay strategy to resolve polytomies, converting them into
    bifurcations.

    For example,
    ```
      1
     /|\
    2 3 4
    ```
    becomes
    ```
      1
     / \
    2   5
       / \
      3   4
    ```

    No adjustments to any branch length columns in `phylogeny_df` are
    performed. However, `origin_time` (as well as all other columns) of a
    polytomy's parent node are duplicated in splayed-out nodes that resolve
    that polytomy. So, nodes added to perform the splaying-out will have zero-
    length subtending branches in this regard (i.e., their origin time will
    match their parent's).

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.
    """

    if "branch_length" in phylogeny_df or "edge_length" in phylogeny_df:
        warnings.warn(
            "alifestd_splay_polytomies does not update branch length columns. "
            "Use `origin_time` to recalculate branch lengths for collapsed "
            "phylogeny."
        )

    assert alifestd_is_asexual(phylogeny_df)

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    splay_to_lookup = dict()
    # splay_parent_lookup = {id_: id_ for id_ in phylogeny_df["id"]}
    splay_from_lookup = dict()
    phylogeny_df.set_index("id", drop=False, inplace=True)

    new_id_generator = it.count(phylogeny_df["id"].max() + 1)

    new_rows = dict()
    for id_, row in phylogeny_df.iterrows():
        ancestor_id = row["ancestor_id"]
        if ancestor_id == id_:  # root
            continue

        if ancestor_id not in splay_to_lookup:
            splayed_from_id = ancestor_id
            splayed_to_id = ancestor_id
        else:
            splayed_from_id = splay_to_lookup[ancestor_id]
            splayed_to_id = next(new_id_generator)

            phylogeny_df.loc[id_, "ancestor_id"] = splayed_to_id
            new_row = phylogeny_df.loc[ancestor_id].copy()
            new_row["id"] = splayed_to_id
            new_row["ancestor_id"] = splayed_from_id
            new_rows[splayed_to_id] = new_row

        splay_from_lookup[splayed_to_id] = splayed_from_id
        splay_to_lookup[ancestor_id] = splayed_to_id

    # drop unnecessary splays
    dropped_splay_remap = dict()
    for ancestor_id, splayed_to_id in splay_to_lookup.items():
        if ancestor_id != splayed_to_id:
            splayed_from_id = splay_from_lookup[splayed_to_id]
            assert new_rows[splayed_to_id]["ancestor_id"] == splayed_from_id
            dropped_splay_remap[splayed_to_id] = splayed_from_id
            del new_rows[splayed_to_id]

    # correct dropped splays
    phylogeny_df["ancestor_id"] = (
        phylogeny_df["ancestor_id"]
        .map(
            {
                **{
                    ancestor_id: ancestor_id
                    for ancestor_id in phylogeny_df["ancestor_id"]
                },
                **dropped_splay_remap,
            }
        )
        .astype(int)
    )

    appendum_df = pd.DataFrame(new_rows.values())
    phylogeny_df = pd.concat([phylogeny_df, appendum_df], ignore_index=True)

    phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        phylogeny_df["id"], phylogeny_df["ancestor_id"]
    )

    return phylogeny_df.reset_index(drop=True)



import typing

from ....._auxiliary_lib import is_nondecreasing
from ._iter_rank_backstops import iter_rank_backstops
from ._iter_rank_seps import iter_rank_seps
from ._iter_target_ranks import iter_target_ranks


def get_retained_ranks(
    degree: int,
    interspersal: int,
    num_strata_deposited: int,
) -> typing.Set[int]:
    """Calculate the set of strata ranks retained at `num_strata_deposited`."""
    # special case
    if num_strata_deposited == 0:
        return set()

    last_rank = num_strata_deposited - 1
    # we will always retain the zeroth rank and the last rank
    # Set data structure prevents duplicates
    res = {0, last_rank}

    for target_rank, rank_backstop, retained_ranks_sep in zip(
        iter_target_ranks(degree, num_strata_deposited),
        iter_rank_backstops(degree, interspersal, num_strata_deposited),
        iter_rank_seps(degree, interspersal, num_strata_deposited),
    ):

        min_retained_rank = rank_backstop

        target_ranks = range(
            min_retained_rank,  # start
            num_strata_deposited,  # stop, not inclusive
            retained_ranks_sep,  # sep
        )

        # sanity checks
        # ensure target_ranks non-empty
        assert len(target_ranks)
        # ensure expected ordering of target ranks
        assert is_nondecreasing(target_ranks)
        # ensure last coverage at or past the target
        assert target_ranks[0] <= target_rank
        # ensure one-past-midpoint coverage before the target
        if len(target_ranks) >= 3:
            assert target_ranks[len(target_ranks) // 2 + 1] > target_rank
        # ensure at least interspersal ranks covered
        try:
            assert len(target_ranks) >= min(
                interspersal,
                len(range(target_rank, num_strata_deposited)),
            )
        except OverflowError:
            pass
        # ensure space complexity cap respected
        assert len(target_ranks) <= 2 * (interspersal + 1)
        # ensure sufficient target_ranks included
        if retained_ranks_sep > 1:
            assert len(target_ranks) >= interspersal

        # add to retained set
        res.update(target_ranks)

    # sanity checks then return
    assert all(0 <= n < num_strata_deposited for n in res)
    assert res
    return res



from ..._impl import CalcMrcaUncertaintyAbsUpperBoundWorstCase


# must inherit rather than assign due to failure of attribute lookup when
# pickling otherwise
class CalcMrcaUncertaintyAbsUpperBound(
    CalcMrcaUncertaintyAbsUpperBoundWorstCase
):
    pass



import typing

import numpy as np
import pandera as pa

from .._auxiliary_lib import CopyableSeriesItemsIter

_unsigned_integer_series_t = typing.Union[
    pa.typing.Series[pa.typing.UInt8()],
    pa.typing.Series[pa.typing.UInt16()],
    pa.typing.Series[pa.typing.UInt32()],
    pa.typing.Series[pa.typing.UInt64()],
]


class HereditaryStratigraphicSpecimen:
    """Postprocessing representation of the differentia retained by an extant
    HereditaryStratigraphicColumn, indexed by deposition rank.

    All entries correspond to retained differentia (i.e., no entries are null).

    See Also
    --------
    HereditaryStratigraphicAssemblageSpecimen
        Specimen representation that allows for easier alignment among members
        of a population without perfectly homogeneous retained ranks.
    specimen_from_records
        Deserialize a `HereditaryStratigraphicSpecimen` from a dict composed of
        builtin data types.
    col_to_specimen
        Create a `HereditaryStratigraphicSpecimen` from a
        `HereditaryStratigraphicColumn`.
    """

    __slots__ = ("_data", "_stratum_differentia_bit_width")

    _data: _unsigned_integer_series_t
    _stratum_differentia_bit_width: int

    def __init__(
        self: "HereditaryStratigraphicSpecimen",
        stratum_differentia_series: _unsigned_integer_series_t,
        stratum_differentia_bit_width: int,
    ) -> None:
        """Initialize a HereditaryStratigraphicSpecimen object with a sequence
        of rank-indexed differentia and a differentia bit width."""
        self._data = stratum_differentia_series
        self._stratum_differentia_bit_width = stratum_differentia_bit_width

        assert self._data.index.dtype in (
            "int8",
            "uint8",
            "int16",
            "uint16",
            "int32",
            "uint32",
            "int64",
            "uint64",
        )
        self._data.index.astype(np.uint64, copy=False)

    def GetStratumDifferentiaBitWidth(
        self: "HereditaryStratigraphicSpecimen",
    ) -> int:
        """How many bits wide are the differentia of strata?"""
        return self._stratum_differentia_bit_width

    def GetNumStrataDeposited(
        self: "HereditaryStratigraphicSpecimen",
    ) -> int:
        """How many strata have been depostited on the column?

        Note that a first stratum is deposited on the column during
        initialization.
        """
        assert self._data.index.dtype in (
            "int8",
            "uint8",
            "int16",
            "uint16",
            "int32",
            "uint32",
            "int64",
            "uint64",
        )
        # self._data.index[-1] is integral
        # must convert to int, or (at least sometimes) float results after + 1
        return int(self._data.index[-1]) + 1

    def GetNumStrataRetained(
        self: "HereditaryStratigraphicSpecimen",
    ) -> int:
        """How many strata are currently stored within the column?

        May be fewer than the number of strata deposited if strata have been
        discarded as part of the configured stratum retention policy.
        """
        return len(self._data)

    def GetNumDiscardedStrata(
        self: "HereditaryStratigraphicSpecimen",
    ) -> int:
        """How many deposited strata have been discarded?

        Determined by number of generations elapsed and the configured column
        retention policy.
        """
        return self.GetNumStrataDeposited() - self.GetNumStrataRetained()

    def HasDiscardedStrata(
        self: "HereditaryStratigraphicSpecimen",
    ) -> bool:
        """Have any deposited strata been discarded?"""
        return self.GetNumDiscardedStrata() > 0

    def GetData(
        self: "HereditaryStratigraphicSpecimen",
    ) -> _unsigned_integer_series_t:
        """Get the underlying Pandas Series containing differentia values
        indexed by rank.

        Notes
        -----
        This function directly returns the specimen's underlying Series
        data, so mutation of the returned object will alter or invalidate this
        specimen.
        """
        return self._data

    def GetDifferentiaVals(
        self: "HereditaryStratigraphicSpecimen",
    ) -> np.ndarray:
        """Get the integer underlying values in the stored Pandas Series.

        Returns
        -------
        differentia : np.ndarray
            A 1-dimensional NumPy integer ndarray containing differentia values,
            including garbage values where the underlying Series is null.

        Notes
        -----
        This function returns a direct view into the Series data, so no copy is
        made. Changes to the returned array will propagate to the Series
        object's underlying values, and vice versa.
        """
        return self._data.array.to_numpy(copy=False)

    def GetRankIndex(
        self: "HereditaryStratigraphicSpecimen",
    ) -> np.ndarray:
        """Get the integer index in the stored Pandas Series, representing the
        ranks of stratum entries.

        Returns
        -------
        ranks : np.ndarray
            A numpy array containing ranks of differentia entries, including
            null entries for differentia that are not retained.
        """
        return self._data.index.array.to_numpy(copy=False, dtype=np.uint64)

    def GetRankAtColumnIndex(
        self: "HereditaryStratigraphicSpecimen",
        index: int,
    ) -> int:
        """Map array position to generation of deposition.

        What is the deposition rank of the stratum positioned at index i
        among retained strata? Index order is from most ancient (index 0) to
        most recent.
        """
        return self.GetRankIndex()[index]

    def IterRetainedRanks(
        self: "HereditaryStratigraphicSpecimen",
    ) -> typing.Iterator[int]:
        """Iterate over deposition ranks of strata retained in the specimen."""
        yield from self.GetRankIndex()

    def IterRetainedDifferentia(
        self: "HereditaryStratigraphicSpecimen",
    ) -> typing.Iterator[int]:
        """Iterate over differentia of strata retained in the specimen.

        Differentia yielded from most ancient to most recent.
        """
        yield from self.GetDifferentiaVals()

    def IterRankDifferentiaZip(
        self: "HereditaryStratigraphicSpecimen",
        copyable: bool = False,
    ) -> typing.Iterator[typing.Tuple[int, int]]:
        """Iterate over ranks of retained strata and their differentia.

        If `copyable`, return an iterator that can be copied to produce a new
        fully-independent iterator at the same position.

        Equivalent to `zip(specimen.IterRetainedRanks(),
        specimen.IterRetainedDifferentia())`, but may be more efficient.
        """
        if copyable:
            return CopyableSeriesItemsIter(self._data)
        else:
            return self._data.items()



import math

from ._calc_target_rank import calc_target_rank
from ._calc_target_recency import calc_target_recency


def calc_rank_cutoff(
    degree: int,
    interspersal: int,
    pow_: int,
    num_strata_deposited: int,
) -> int:
    """Before what rank should no strata be retained to provide coverage for the `pow`th target recency?

    Will be less than target rank to ensure adequate resolution at target
    rank and just greater than target rank. Inclusive (i.e., this rank may
    be retained). Will monotonically increase with num_strata_deposited.
    """
    target_recency = calc_target_recency(
        degree,
        pow_,
        num_strata_deposited,
    )
    rank_cutoff = max(
        num_strata_deposited
        - int(math.ceil(target_recency * (interspersal + 1) / interspersal)),
        0,
    )
    assert rank_cutoff <= calc_target_rank(
        degree,
        pow_,
        num_strata_deposited,
    )
    if num_strata_deposited == 0:
        assert rank_cutoff == 0
    else:
        assert 0 <= rank_cutoff <= num_strata_deposited - 1
    return rank_cutoff



import itertools as it
import unittest

from hstrat._auxiliary_lib import all_same


class TestAllSame(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        for container in list, dict, set, iter:
            assert all_same(container(()))

    def test_singleton(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert all_same(container((content,)))

        for container in (set,):
            for hashable_content in None, True, False, "a", 1, 0, "asdf":
                assert all_same(container((hashable_content,)))

    def test_doublet_same(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert all_same(container((content, content)))

        for container in (set,):
            for hashable_content in None, True, False, "a", 1, 0, "asdf":
                assert all_same(
                    container((hashable_content, hashable_content))
                )

    def test_doublet_notsame(self):
        content = None, True, False, "a", 1, 0, "asdf", [1, 2, 3]
        for container in list, iter:
            for a, b in it.permutations(content, 2):
                assert all_same(container((a, b))) == (a == b)

        hashable_content = None, True, False, "a", 1, 0, "asdf"
        for container in (set,):
            for a, b in it.permutations(hashable_content, 2):
                assert all_same(container((a, b))) == (a == b)

    def test_triplet_same(self):
        for container in list, iter:
            for content in None, True, False, "a", 1, 0, "asdf", [1, 2, 3]:
                assert all_same(container((content, content, content)))

        for container in (set,):
            for hashable_content in None, True, False, "a", 1, 0, "asdf":
                assert all_same(
                    container(
                        (
                            hashable_content,
                            hashable_content,
                            hashable_content,
                        )
                    )
                )

    def test_triplet_onenotsame(self):
        content = None, True, False, "a", 1, 0, "asdf", [1, 2, 3]
        for container in list, iter:
            for a, b in it.permutations(content, 2):
                assert all_same(container((a, b, a))) == (a == b)
                assert all_same(container((a, a, b))) == (a == b)

        hashable_content = None, True, False, "a", 1, 0, "asdf"
        for container in (set,):
            for a, b in it.permutations(hashable_content, 2):
                assert all_same(container((a, b, a))) == (a == b)
                assert all_same(container((a, a, b))) == (a == b)

    def test_triplet_allnotsame(self):
        content = None, True, False, "a", 1, 0, "asdf", [1, 2, 3]
        for container in list, iter:
            for a, b, c in it.permutations(content, 3):
                assert all_same(container((a, b, c))) == (a == b == c)

        hashable_content = None, True, False, "a", 1, 0, "asdf"
        for container in (set,):
            for a, b, c in it.permutations(hashable_content, 3):
                assert all_same(container((a, b, c))) == (a == b == c)


if __name__ == "__main__":
    unittest.main()



import typing

from deprecated.sphinx import deprecated

from ..genome_instrumentation import HereditaryStratum
from ._pack_differentiae_str import pack_differentiae_str


@deprecated(
    version="1.8.0",
    reason="Use pack_differentiae_str instead.",
)
def pack_differentiae(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> str:
    """Pack a sequence of differentiae together into a compact
    representation.

    Returns a string with base 64 encoded concatenation of diffferentiae.
    If `differentia_bit_width` is not an even byte multiple, the first encoded
    byte tells how many empty padding bits, if any, were placed at the end of
    the concatenation in order to align the bitstring end to byte boundaries.
    """

    return pack_differentiae_str(strata, differentia_bit_width)



import itertools as it
import typing


# adapted from https://docs.python.org/3/library/itertools.html
def pairwise(iterable: typing.Iterable) -> typing.Iterator:
    """Return successive overlapping pairs taken from the input iterable.

    The number of 2-tuples in the output iterator will be one fewer than the
    number of inputs. It will be empty if the input iterable has fewer than two
    values.
    """
    # pairwise('ABCDEFG') --> AB BC CD DE EF FG
    a, b = it.tee(iterable)
    next(b, None)
    return zip(a, b)


# use library implementation if available (Python 3.10+)
try:
    pairwise = it.pairwise  # noqa: F811
except AttributeError:
    pass



from stratum_retention_algorithms import (
    UnsatisfiableParameterizationRequestError,
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    provided_stratum_retention_algorithms,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    recency_proportional_resolution_curbed_algo,
    stochastic_algo,
)
from stratum_retention_policy_evaluators import (
    MrcaUncertaintyAbsExactEvaluator,
    MrcaUncertaintyAbsUpperBoundEvaluator,
    MrcaUncertaintyRelExactEvaluator,
    MrcaUncertaintyRelUpperBoundEvaluator,
    NumStrataRetainedExactEvaluator,
    NumStrataRetainedUpperBoundEvaluator,
)
from stratum_retention_policy_parameterizers import (
    PropertyAtLeastParameterizer,
    PropertyAtMostParameterizer,
    PropertyExactlyParameterizer,
)

from . import (
    stratum_retention_algorithms,
    stratum_retention_policy_evaluators,
    stratum_retention_policy_parameterizers,
)

__all__ = [
    "stratum_retention_policy_parameterizers",
    "stratum_retention_algorithms",
    "stratum_retention_policy_evaluators",
    "MrcaUncertaintyAbsExactEvaluator",
    "MrcaUncertaintyAbsUpperBoundEvaluator",
    "MrcaUncertaintyRelExactEvaluator",
    "MrcaUncertaintyRelUpperBoundEvaluator",
    "NumStrataRetainedExactEvaluator",
    "NumStrataRetainedUpperBoundEvaluator",
    "UnsatisfiableParameterizationRequestError",
    "depth_proportional_resolution_algo",
    "depth_proportional_resolution_tapered_algo",
    "fixed_resolution_algo",
    "geom_seq_nth_root_algo",
    "geom_seq_nth_root_tapered_algo",
    "nominal_resolution_algo",
    "perfect_resolution_algo",
    "provided_stratum_retention_algorithms",
    "pseudostochastic_algo",
    "recency_proportional_resolution_algo",
    "recency_proportional_resolution_curbed_algo",
    "stochastic_algo",
    "PropertyAtLeastParameterizer",
    "PropertyAtMostParameterizer",
    "PropertyExactlyParameterizer",
]



import typing

from ...._detail import PolicyCouplerBase
from ...._impl import GenDropRanksFromPredKeepRank
from ..._PolicySpec import PolicySpec


class _PredKeepRank:
    """Functor to implement the nominal resolution stratum retention policy, for
    use with HereditaryStratigraphicColumn.

    This functor enacts the nominal resolution policy by specifying the set of
    strata ranks that should be purged from a hereditary stratigraphic column
    when the nth stratum is deposited.
    """

    def __init__(
        self: "_PredKeepRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "_PredKeepRank", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "_PredKeepRank",
        policy: typing.Optional[PolicyCouplerBase],
        num_stratum_depositions_completed: typing.Optional[int],
        stratum_rank: typing.Optional[int],
    ) -> bool:
        """Decide if a stratum within the stratagraphic column should be
        retained or purged.

        Every time a new stratum is deposited, this method is called on each
        stratum present in a HereditaryStratigraphicColumn to determine whether
        it should be retained. Strata that return False are immediately purged
        from the column, meaning that for a stratum to persist it must earn a
        True result from this method each and every time a new stratum is
        deposited.

        Parameters
        ----------
        policy: Policy
            Policy this functor enacts.
        stratum_rank : int
            The number of strata that were deposited before the stratum under
            consideration for retention.
        num_stratum_depositions_completed : int
            The number of strata that have already been deposited, not
            including the latest stratum being deposited which prompted the
            current purge operation.

        Returns
        -------
        bool
            True if the stratum should be retained, False otherwise.

        See Also
        --------
        nominal_resolution_algo:
            For details on the rationale, implementation, and guarantees of the
            nominal resolution stratum retention policy.
        """
        return stratum_rank in (0, num_stratum_depositions_completed)


FromPredKeepRank = GenDropRanksFromPredKeepRank(_PredKeepRank)



import pandas as pd
import pytest

from hstrat import hstrat
from hstrat.juxtaposition._impl._dispatch_impl import dispatch_impl


def test_dispatch_impl_column():
    target = hstrat.HereditaryStratigraphicColumn()
    impl_module = dispatch_impl(target, target)
    assert impl_module.__name__.endswith("_impl_column")


def test_dispatch_impl_specimen():
    target = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series([], dtype="int8"), 8
    )
    impl_module = dispatch_impl(target, target)
    assert impl_module.__name__.endswith("_impl_specimen")


def test_dispatch_impl_invalid_input():
    target = "invalid"
    with pytest.raises(TypeError):
        dispatch_impl(target, target)



import os

import polars as pl

from hstrat._auxiliary_lib import (
    alifestd_try_add_ancestor_list_col,
    alifestd_validate,
)
from hstrat.dataframe import surface_unpack_reconstruct

assets_path = os.path.join(os.path.dirname(__file__), "assets")


def test_smoke():
    df = pl.read_csv(f"{assets_path}/packed.csv")
    res = surface_unpack_reconstruct(df)
    assert len(res) > len(df)
    assert alifestd_validate(
        alifestd_try_add_ancestor_list_col(res.to_pandas()),
    )



"""Tools to estimate MRCA generation from shared ranks of commonality and
disparity."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(
    __name__, __file__, should_launder=lambda _: False
)
del lazy_attach_stub



import typing

from iterpop import iterpop as ip
import opytional as opyt

from ...._auxiliary_lib import HereditaryStratigraphicArtifact
from ....juxtaposition._impl import dispatch_impl


def extract_common_retained_ranks_through_first_retained_disparity(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
) -> typing.List[int]:
    """Extracts a list of common retained ranks between two hereditary
    stratigraphic artifacts up to and including the rank of the first retained
    disparity.

    If no disparity exists (i.e., columns are exactly identical), the rank
    which is next to be deposited will be considered the next disparity.
    """
    # choose correct impl for columns/specimens
    impl = dispatch_impl(first, second)

    rank_of_first_retained_disparity_between = opyt.or_else(
        impl.calc_rank_of_first_retained_disparity_between(
            first,
            second,
            confidence_level=0.49,
        ),
        lambda: ip.pophomogeneous(
            (
                first.GetNumStrataDeposited(),
                second.GetNumStrataDeposited(),
            )
        ),
    )
    return [
        *impl.iter_ranks_of_retained_commonality_between(first, second),
        rank_of_first_retained_disparity_between,
    ]



import numbers
import typing

import networkx as nx
import numpy as np
import sortedcontainers as sc


def estimate_origin_times(
    graph: nx.Graph,
    leaf_node_origin_times: typing.Dict[typing.Any, numbers.Number],
) -> typing.Dict[str, numbers.Number]:
    """Estimate origin time of graph nodes by working backward from
    phylogenetic depths of leaf nodes using "length" attributes on graph
    edges.

    Keys in `leaf_node_origin_times` should correspond to nodes in graph.
    """

    assert all(data["length"] >= 0 for _a, _b, data in graph.edges(data=True))

    # ensure clean slate for origin time setup
    for _node, data in graph.nodes(data=True):
        if "origin_time" in data:
            del data["origin_time"]

    # setup origin times of leaf nodes in graph
    leaf_nodes = leaf_node_origin_times.keys()
    for leaf_node in leaf_nodes:
        graph.nodes[leaf_node]["origin_time"] = leaf_node_origin_times[
            leaf_node
        ]

    # starting from leaf nodes...
    visited_nodes = set()  # fully explored nodes
    discovered_nodes = sc.SortedSet(
        leaf_nodes,
        key=lambda node: graph.nodes[node]["origin_time"],
    )  # nodes eligible to visit next, sorted chronologically (deepest last)

    # ... traverse graph: discover nodes and calculate their origin times
    # in reverse chronological order
    while len(visited_nodes) < len(graph):
        # pop oldest discovered node, mark visited, discover its neighbors
        cur_node = discovered_nodes.pop(-1)
        visited_nodes.add(cur_node)

        newly_discovered_nodes = set(graph.neighbors(cur_node)) - (
            visited_nodes | discovered_nodes
        )
        # enforce chronological consistency:
        # ensure that nodes are only visited after all children have been
        # visited and at most one unvisited neighbor (ancestor) remains
        newly_discovered_nodes -= {
            node
            for node in newly_discovered_nodes
            if sum(
                "origin_time" not in graph.nodes[neighbor]
                for neighbor in graph.neighbors(node)
            )
            > 1
        }
        # give newly discovered nodes origin time estimates
        for newly_discovered_node in newly_discovered_nodes:
            assert "origin_time" not in graph[newly_discovered_node]
            # estimate origin_time of cur node
            descendant_origin_times = [
                graph.nodes[neighbor_node]["origin_time"]
                for neighbor_node in graph.neighbors(newly_discovered_node)
                if "origin_time" in graph.nodes[neighbor_node]
            ]
            estimated_origin_times = [
                graph.nodes[neighbor_node]["origin_time"]
                - graph[newly_discovered_node][neighbor_node]["length"]
                for neighbor_node in graph.neighbors(newly_discovered_node)
                if "origin_time" in graph.nodes[neighbor_node]
            ]
            assert estimated_origin_times
            origin_time = min(
                np.mean(estimated_origin_times),
                min(descendant_origin_times),
            )
            graph.nodes[newly_discovered_node]["origin_time"] = origin_time

        # register newly discovered nodes as eligible to visit next
        discovered_nodes.update(newly_discovered_nodes)

    return {node: graph.nodes[node]["origin_time"] for node in graph.nodes}



import itertools as it
import random

import pytest

from hstrat._auxiliary_lib import all_same
from hstrat.hstrat import HereditaryStratigraphicColumn, stochastic_algo


@pytest.mark.parametrize(
    "impl",
    stochastic_algo._GenDropRanks_impls,
)
@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_impl_stochasticity(impl, replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
    )
    for num_strata_deposited in range(1, 10**3):
        if all_same(
            it.chain(
                (
                    sorted(
                        impl(spec)(
                            policy,
                            num_strata_deposited,
                            column.IterRetainedRanks(),
                        )
                    )
                ),
                (
                    sorted(
                        impl(spec)(
                            policy,
                            num_strata_deposited,
                            column.IterRetainedRanks(),
                        )
                    )
                ),
            )
        ):
            break
        else:
            assert False, "All implementations were identical."


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_eq(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    instance = stochastic_algo.GenDropRanks(spec)

    assert instance == instance
    assert instance == stochastic_algo.GenDropRanks(spec)
    assert instance is not None



from hstrat._auxiliary_lib import is_in


def test_is_in_true():
    data = [1, 2, 3]
    assert is_in(data, [data, [4, 5, 6], [7, 8, 9]])


def test_is_in_false():
    data = [1, 2, 3]
    assert not is_in(data, [[4, 5, 6], [7, 8, 9]])


def test_is_in_empty_iterator():
    data = {"a": 1, "b": 2}
    assert not is_in(data, [])


def test_is_in_identical_objects():
    data = [1, 2, 3]
    assert not is_in(data, [data.copy(), [4, 5, 6], [7, 8, 9]])



import random
import typing

import opytional as opyt

from hstrat import hstrat


class SimpleGenomeAnnotatedWithSparseRetention:

    content: float
    annotation: hstrat.HereditaryStratigraphicColumn

    def __init__(
        self: "SimpleGenomeAnnotatedWithSparseRetention",
        content_: typing.Optional[float] = None,
        annotation_: typing.Optional[
            hstrat.HereditaryStratigraphicColumn
        ] = None,
    ) -> None:
        self.content = opyt.or_value(
            content_,
            random.uniform(0, 100),
        )
        self.annotation = opyt.or_value(
            annotation_,
            hstrat.HereditaryStratigraphicColumn(
                hstrat.fixed_resolution_algo.Policy(25),
            ),
        )

    def CreateOffspring(
        self: "SimpleGenomeAnnotatedWithSparseRetention",
    ) -> "SimpleGenomeAnnotatedWithSparseRetention":
        return SimpleGenomeAnnotatedWithSparseRetention(
            self.content + random.uniform(0.0, 1.0) % 100.0,
            self.annotation.CloneDescendant(),
        )



import itertools as it
import random

import opytional as opyt
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
def test_CalcRankOfFirstRetainedDisparityBetween_specimen(
    retention_policy, differentia_width
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_rank_of_first_retained_disparity_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column)
    ) == hstrat.calc_rank_of_first_retained_disparity_between(column, column)

    assert hstrat.calc_rank_of_first_retained_disparity_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(column2)
    ) == hstrat.calc_rank_of_first_retained_disparity_between(column, column2)

    assert hstrat.calc_rank_of_first_retained_disparity_between(
        hstrat.col_to_specimen(column), hstrat.col_to_specimen(child1)
    ) == hstrat.calc_rank_of_first_retained_disparity_between(column, child1)

    assert hstrat.calc_rank_of_first_retained_disparity_between(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_rank_of_first_retained_disparity_between(child1, child2)

    child1.DepositStrata(10)
    assert hstrat.calc_rank_of_first_retained_disparity_between(
        hstrat.col_to_specimen(child1), hstrat.col_to_specimen(child2)
    ) == hstrat.calc_rank_of_first_retained_disparity_between(child1, child2)


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_asynchronous(
    retention_policy,
    ordered_store,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_rank_of_first_retained_disparity_between(
                first,
                second,
            ) == hstrat.calc_rank_of_first_retained_disparity_between(
                second,
                first,
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreDict,
            marks=pytest.mark.heavy_2b,
        ),
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_commutativity_synchronous(
    retention_policy,
    ordered_store,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_rank_of_first_retained_disparity_between(
                first,
                second,
            ) == hstrat.calc_rank_of_first_retained_disparity_between(
                second,
                first,
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_comparison_validity(retention_policy, ordered_store):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_ordered_store=ordered_store,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for generation in range(100):
        for first, second in it.combinations(population, 2):
            lcrw = hstrat.calc_rank_of_last_retained_commonality_between(
                first, second
            )
            if lcrw is not None:
                assert 0 <= lcrw <= generation

            fdrw = hstrat.calc_rank_of_first_retained_disparity_between(
                first,
                second,
            )
            if fdrw is not None:
                assert 0 <= fdrw <= generation

            assert hstrat.calc_rank_of_mrca_bounds_between(
                first, second, prior="arbitrary"
            ) in [
                (lcrw, opyt.or_value(fdrw, first.GetNumStrataDeposited())),
                None,
            ]
            if lcrw is not None and fdrw is not None:
                assert lcrw < fdrw

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy1",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "retention_policy2",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_mrca(
    retention_policy1,
    retention_policy2,
    ordered_store,
):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy1,
    )
    second = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy2,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(first, second)
            == 0
        )
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(first, second)
            == 0
        )
        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_no_divergence(retention_policy, ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for _generation in range(100):
        _ = _generation
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                column, column
            )
            is None
        )
        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2,
        ),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_single_branch_divergence(retention_policy, ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    for generation in range(40):
        branch_column = column.CloneDescendant()
        for _branch_generation in range(40):
            _ = _branch_generation
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    column, branch_column
                )
                == generation + 1
            )
            branch_column.DepositStratum()

        column.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2,
        ),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_scenario_partial_even_divergence(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    first.DepositStrata(100)

    second = first.Clone()

    first.DepositStratum()
    second.DepositStratum()

    for generation in range(101, 200):
        assert (
            100
            <= hstrat.calc_rank_of_first_retained_disparity_between(
                first, second
            )
            <= generation
        )

        first.DepositStratum()
        second.DepositStratum()


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        pytest.param(
            hstrat.HereditaryStratumOrderedStoreTree,
            marks=pytest.mark.heavy,
        ),
    ],
)
def test_scenario_partial_uneven_divergence(retention_policy, ordered_store):
    first = hstrat.HereditaryStratigraphicColumn(
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=retention_policy,
    )

    first.DepositStrata(100)

    second = first.Clone()

    first.DepositStratum()

    for generation in range(101, 200):
        assert (
            100
            <= hstrat.calc_rank_of_first_retained_disparity_between(
                first, second
            )
            <= generation
        )
        first.DepositStratum()

    second.DepositStratum()

    for generation in range(101, 200):
        assert (
            100
            <= hstrat.calc_rank_of_first_retained_disparity_between(
                first, second
            )
            <= generation + 1
        )
        second.DepositStratum()


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8],
)
def test_CalcRankOfFirstRetainedDisparityWith1(
    ordered_store,
    differentia_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
        stratum_ordered_store=ordered_store,
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    if differentia_width == 64:
        for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c2,
                )
                == column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited()
            )

    elif differentia_width == 1:
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                offspring1,
                offspring2,
            )
            < offspring2.GetNumStrataDeposited() - 1
        )
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                offspring2,
                offspring1,
            )
            < offspring2.GetNumStrataDeposited() - 1
        )
        for c in [offspring1, offspring2]:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c,
                    column,
                )
                < column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    column,
                    c,
                )
                < column.GetNumStrataDeposited()
            )

    for c in [column, offspring1, offspring2]:
        if (
            c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.2,
            )
            > 1
        ):
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c,
                    c,
                    0.8,
                )
                == c.GetNumStrataDeposited()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=0.2,
                )
                + 1
            )
        else:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c,
                    c,
                    0.8,
                )
                is None
            )
        if (
            c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.05,
            )
            > 1
        ):
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c,
                    c,
                    0.95,
                )
                == c.GetNumStrataDeposited()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=0.05,
                )
                + 1
            )
        else:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c,
                    c,
                    0.95,
                )
                is None
            )

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c1,
                    c2,
                )
                == column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited()
            )
        elif differentia_width == 1:
            assert hstrat.calc_rank_of_first_retained_disparity_between(
                c1, c2, 0.999999
            ) < hstrat.calc_rank_of_first_retained_disparity_between(
                c2, c1, 0.8
            )

    for c in [column, offspring1, offspring2]:
        if (
            c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.2,
            )
            > 1
        ):
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(c, c, 0.8)
                == c.GetNumStrataDeposited()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=0.2,
                )
                + 1
            )
        else:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(c, c, 0.8)
                is None
            )
        if (
            c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.05,
            )
            > 1
        ):
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c, c, 0.95
                )
                == c.GetNumStrataDeposited()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=0.05,
                )
                + 1
            )
        else:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c, c, 0.95
                )
                is None
            )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcRankOfFirstRetainedDisparityWith2(ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )

    while True:
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res == 0
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
            0.49,
        )
        assert res in (None, 1)
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
            0.49,
        )
        if res == 1:
            break

    while True:
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res == 0
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
            0.49,
        )
        assert res in (None, 1)
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
            0.49,
        )
        if res is None:
            break

    for rep in range(500):
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
            0.49,
        )
        assert res in (None, 1)
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
            0.49,
        )

        for gen in range(8):
            offspring1.DepositStratum()
            offspring2.DepositStratum()
            # should occur about 1 in 20 times
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    offspring2,
                    offspring1,
                )
                is not None
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    offspring1,
                    offspring2,
                )
                is not None
            )

    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )

    for __ in range(100):
        column.DepositStratum()

    for rep in range(5):
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        divergence_rank = offspring1.GetNumStrataDeposited() - 1

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res < divergence_rank
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        for gen in range(10):
            offspring1.DepositStratum()
            offspring2.DepositStratum()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res is not None
        assert res < offspring1.GetNumStrataDeposited() - 1
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        for gen in range(42):
            offspring1.DepositStratum()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res is not None
        assert res < offspring2.GetNumStrataDeposited()
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcRankOfFirstRetainedDisparityWith3(ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.nominal_resolution_algo.Policy(),
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    offspring1.DepositStrata(100)
    offspring2.DepositStrata(100)

    for c2 in [offspring1, offspring2]:
        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                column,
                c2,
            )
            == column.GetNumStrataDeposited()
        )

        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                c2,
                column,
            )
            == column.GetNumStrataDeposited()
        )

    assert (
        hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        == offspring1.GetNumStrataDeposited() - 1
    )
    assert (
        hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )
        == offspring1.GetNumStrataDeposited() - 1
    )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
def test_CalcRankOfFirstRetainedDisparityWith4(ordered_store):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    while True:
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res == 0
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        for gen in range(20):
            offspring1.DepositStratum()

        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                offspring2,
                offspring1,
            )
            is not None
        )

        for gen in range(20):
            offspring2.DepositStratum()

        assert (
            hstrat.calc_rank_of_first_retained_disparity_between(
                offspring2,
                offspring1,
            )
            is not None
        )
        break

    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    for __ in range(100):
        column.DepositStratum()

    for rep in range(50):
        offspring1 = column.CloneDescendant()
        offspring2 = column.CloneDescendant()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )

        idx = (
            offspring1.GetNumStrataRetained()
            - offspring1.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                significance_level=0.05
            )
            + 1
        )
        assert (
            offspring1.GetRankAtColumnIndex(idx - 1)
            <= res
            <= offspring1.GetRankAtColumnIndex(idx)
        )

        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        for gen in range(10):
            offspring1.DepositStratum()
            offspring2.DepositStratum()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res is not None
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )

        for gen in range(42):
            offspring1.DepositStratum()

        res = hstrat.calc_rank_of_first_retained_disparity_between(
            offspring1,
            offspring2,
        )
        assert res is not None
        assert res == hstrat.calc_rank_of_first_retained_disparity_between(
            offspring2,
            offspring1,
        )


@pytest.mark.parametrize(
    "ordered_store",
    [
        hstrat.HereditaryStratumOrderedStoreDict,
        hstrat.HereditaryStratumOrderedStoreList,
        hstrat.HereditaryStratumOrderedStoreTree,
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8],
)
def test_CalcRankOfFirstRetainedDisparityWith5(
    ordered_store,
    differentia_width,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_width,
        stratum_ordered_store=ordered_store,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    column.DepositStrata(100)

    offspring1 = column.CloneDescendant()
    offspring2 = column.CloneDescendant()

    assert column.GetNumStrataDeposited() == 101
    assert offspring1.GetNumStrataDeposited() == 102
    assert offspring2.GetNumStrataDeposited() == 102

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c1,
                    c2,
                )
                == column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited()
            )
        elif differentia_width == 1:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c1,
                    c2,
                )
                < column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c1,
                )
                < column.GetNumStrataDeposited()
            )

    for c in [column, offspring1, offspring2]:
        for conf in 0.8, 0.95, 0.99:
            col_idx = (
                c.GetNumStrataRetained()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=1 - conf,
                )
                + 1
            )
            if col_idx == c.GetNumStrataRetained():
                assert (
                    c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                        significance_level=1 - conf,
                    )
                    == 1
                )
                assert (
                    hstrat.calc_rank_of_first_retained_disparity_between(
                        c, c, conf
                    )
                    is None
                )
            else:
                assert hstrat.calc_rank_of_first_retained_disparity_between(
                    c, c, conf
                ) == c.GetRankAtColumnIndex(col_idx)

    offspring1.DepositStrata(99)
    offspring2.DepositStrata(99)

    for c1, c2 in it.combinations([column, offspring1, offspring2], 2):
        if differentia_width == 64:
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c1,
                    c2,
                )
                == column.GetNumStrataDeposited()
            )
            assert (
                hstrat.calc_rank_of_first_retained_disparity_between(
                    c2,
                    c1,
                )
                == column.GetNumStrataDeposited()
            )
        elif differentia_width == 1:
            assert hstrat.calc_rank_of_first_retained_disparity_between(
                c1, c2, 0.999999
            ) < hstrat.calc_rank_of_first_retained_disparity_between(
                c2, c1, 0.8
            )

    for c in [column, offspring1, offspring2]:
        for conf in 0.8, 0.95, 0.99:
            col_idx = (
                c.GetNumStrataRetained()
                - c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                    significance_level=1 - conf,
                )
                + 1
            )
            if col_idx == c.GetNumStrataRetained():
                assert (
                    c.CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
                        significance_level=1 - conf,
                    )
                    == 1
                )
                assert (
                    hstrat.calc_rank_of_first_retained_disparity_between(
                        c, c, conf
                    )
                    is None
                )
            else:
                assert hstrat.calc_rank_of_first_retained_disparity_between(
                    c, c, conf
                ) == c.GetRankAtColumnIndex(col_idx)


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_rank_of_first_retained_disparity_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.calc_rank_of_first_retained_disparity_between(a, b)



import typing


class GetAttrLaunderShim:
    """A functor shim to launder the __module__ attribute of an object to a
    specified module name. This is used to hide implementation details in the
    documentation.

    See Also
    --------
    lazy_attach

    lazy_attach_stub
    """

    _wrapped_getattr: typing.Callable
    _should_launder: typing.Callable
    _module_name: str

    def __init__(
        self: "GetAttrLaunderShim",
        wrapped_getattr: typing.Callable,
        should_launder: typing.Callable,
        module_name: str,
    ) -> None:
        """Initialize functor."""
        self._wrapped_getattr = wrapped_getattr
        self._should_launder = should_launder
        self._module_name = module_name

    def __call__(
        self: "GetAttrLaunderShim", name: str, *args, **kwargs
    ) -> object:
        """Wraps the __getattr__ function to launder the __module__ attribute
        of returned objects."""
        attr = self._wrapped_getattr(name, *args, **kwargs)
        if self._should_launder(name):
            try:
                attr.__module__ = self._module_name
            except (AttributeError, TypeError):
                pass  # module attr not settable for all object types
        return attr



"""Tools to specify stratum retention policy."""

from hstrat._auxiliary_lib import lazy_attach

from . import (
    stratum_retention_algorithms,
    stratum_retention_policy_evaluators,
    stratum_retention_policy_parameterizers,
)

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=[
        "stratum_retention_algorithms",
        "stratum_retention_policy_evaluators",
        "stratum_retention_policy_parameterizers",
    ],
    submod_attrs={
        "stratum_retention_algorithms": stratum_retention_algorithms.__all__,
        "stratum_retention_policy_parameterizers": stratum_retention_policy_parameterizers.__all__,
        "stratum_retention_policy_evaluators": stratum_retention_policy_evaluators.__all__,
    },
    should_launder=[].__contains__,
)
del lazy_attach



import typing


# adapted from https://stackoverflow.com/a/434328
def iter_slices(
    len_seq: int,
    chunk_size: int,
    start: int = 0,
) -> typing.Iterator[slice]:
    """Returns an iterator of slices that chunk a sequence of length `len_seq`
    into `chunk_size` parts.

    Parameters
    ----------
    len_seq : int
        The length of the sequence to be chunked.
    chunk_size : int
        The size of each chunk in the sequence.
    start : int, optional
        The starting position of the first slice in the sequence. Defaults to 0.

    Returns
    -------
    typing.Iterator[slice]
        An iterator of slices representing the chunks of the sequence.

    Examples
    --------
    >>> list(iter_slices(10, 5))
    [slice(0, 5), slice(5, 10)]
    >>> list(iter_slices(10, 5, 2))
    [slice(2, 7), slice(7, 12)]
    >>> list(iter_slices(10, 10))
    [slice(0, 10)]
    """
    return (
        slice(pos, pos + chunk_size)
        for pos in range(start, len_seq, chunk_size)
    )



from __future__ import annotations

from random import randint
import typing

from hstrat import hstrat
from hstrat.genome_instrumentation import HereditaryStratigraphicColumn


class Genome:
    __slots__ = ["annotation", "data"]

    def __init__(
        self,
        data: typing.Optional[typing.List[int]] = None,
        annotation: typing.Optional[HereditaryStratigraphicColumn] = None,
    ) -> None:
        self.data = data or [randint(1, 100)]
        self.annotation = annotation or HereditaryStratigraphicColumn(
            stratum_retention_policy=hstrat.recency_proportional_resolution_algo.Policy(
                8
            ),
            stratum_differentia_bit_width=8,
        )

    def get_descendant(self) -> Genome:
        return Genome(
            self.data + [randint(1, 100)], self.annotation.CloneDescendant()
        )

    @property
    def score(self) -> int:  # weighted average, more recent preferred
        n = len(self.data)
        return sum(sum(self.data[i:]) for i in range(n)) * 2 // (n * (n - 1))


def simulate_evolution(
    parents: typing.List[Genome], *, generations: int, carrying_capacity: int
) -> typing.List[Genome]:
    for _ in range(generations):
        children = sum(
            [[p.get_descendant() for p in parents] for _ in range(3)], []
        )
        children.sort(key=lambda x: x.score)
        parents = children[-carrying_capacity:]
    return parents
